@@ -35,14 +35,52 @@ class Entity:
             self.center = solid.Center()
             self.volume = solid.Volume()
 
-    def similar(self,center:tuple=(0,0,0),volume:float=1,bb:tuple=(0,0,0),
-            tolerance=1e-3)->bool:
+    def similarity(self,center:tuple=(0,0,0),bb:tuple=(0,0,0),volume:float=1,
+            tolerance=1e-2)->bool:
+        """method returns a value for the similary between the entity and the 3 parameters
+           cms, bb, and volume"""
+        cms_rel_dist= np.linalg.norm([self.center.x-center[0], self.center.y-center[1], self.center.z-center[2]])/np.linalg.norm(center)
+        bb_rel_dist=np.linalg.norm([self.bb.xlen-bb[0],self.bb.ylen-bb[1],self.bb.zlen-bb[2]])/np.linalg.norm(bb)
+        vol_rel_dist=np.abs(self.volume-volume)/volume
+        return cms_rel_dist + bb_rel_dist + vol_rel_dist
+
+    def similar(self,center:tuple=(0,0,0),bb:tuple=(0,0,0),volume:float=1,
+            tolerance=1e-2)->bool:
         """method checks whether the entity can be regard as similar with another entities parameters"""
-        cms_close=np.norm(self.center-center)<tolerance
-        bb_close=np.norm(self.bb-bb)<tolerance
-        vol_close=np.abs(self.volume-volume)<tolerance
+        print(center,bb,volume)
+        print([self.center.x, self.center.y, self.center.z],[self.bb.xlen,self.bb.ylen,self.bb.zlen],self.volume)
+        cms_close=np.linalg.norm([self.center.x-center[0], self.center.y-center[1], self.center.z-center[2]])/np.linalg.norm(center)<tolerance
+        bb_close=np.linalg.norm([self.bb.xlen-bb[0],self.bb.ylen-bb[1],self.bb.zlen-bb[2]])/np.linalg.norm(bb)<tolerance 
+        vol_close=np.abs(self.volume-volume)/volume<tolerance
         return (cms_close and bb_close and vol_close)
 
+
+def idx_similar(entity_list,center,bounding_box,volume):
+    """returns the index in the solid_list for which a solid is similar in terms of bounding box, cms, and volume
+       If no similar object is found return -1. 
+    """
+    idx_found=[]
+    found=False
+    for i,ent in enumerate(entity_list):
+      if ent.similar([center.x,center.y,center.z],[bounding_box.xlen,bounding_box.ylen, bounding_box.zlen],volume, tolerance=1e-1):
+        found=True
+        idx_found.append(i)
+    if(len(idx_found)>1):
+      #we have multiple matches - pick the best one
+      dsmall=1e9
+      for i,ent in enumerate([entity_list[idx] for idx in idx_found]):
+        d=ent.similarity([center.x,center.y,center.z],[bounding_box.xlen,bounding_box.ylen, bounding_box.zlen],volume)
+        if(d<dsmall):
+          dsmall=d
+          end_idx=idx_found[i]
+    elif(len(idx_found)==1):
+        end_idx=idx_found[0]
+        print(f'INFO: Found index at {end_idx}')
+    else:
+        end_idx=-1
+        print('INFO: No similar object found')
+    return end_idx
+
 class Assembly:
     """This class encapsulates a set of geometries defined by step-files
     addtionally it provides access to meshing-utilities, and export to a DAGMC-enabled
@@ -126,7 +164,7 @@ class Assembly:
         if(tags_set!=len(self.entities)):
            print(f"WARNING: {len(self.entities)-tags_set} volumes were tagged with the default ({default_tag}) material.")
 
-    def load_stp_file(self,filename: str, scale_factor: float = 1.0):
+    def load_stp_file(self,filename: str, scale_factor: float = 0.1):
         """Loads a stp file and makes the 3D solid and wires available for use.
 
         Args:
@@ -134,7 +172,8 @@ class Assembly:
             scale_factor: a scaling factor to apply to the geometry that can be
                 used to increase the size or decrease the size of the geometry.
                 Useful when converting the geometry to cm for use in neutronics
-                simulations.
+                simulations. The default (0.1) corresponds to the standard setting of OpenCASCADE
+                which assumes mm to the the standard of OpenMC which is cm.
 
         Returns:
             CadQuery.solid
@@ -240,15 +279,18 @@ class Assembly:
         #For now this is a hack relying on the fact that merged is a compund object.
 
         msolids=self.merged.Solids()
+        offset=0
         if(len(msolids)!=len(self.entities)):
-            print("ERROR: the number of merged solids does not match the original number")
-            return
+            print("WARNING: the number of merged solids does not match the original number")
+            offset=len(msolids)-len(self.entities)
+            if(offset<0):
+                offset=0
         for i,s in enumerate(msolids):
             j=i+1
             filename=f"volume_{j}.stl"
-            cq.exporters.export(s,filename,exportType="STL",tolerance=tolerance,angularTolerance=angular_tolerance)
+            status=cq.exporters.export(s,filename,exportType="STL",tolerance=tolerance,angularTolerance=angular_tolerance)
             if(self.verbose>1):
-                print(f"INFO: export to file {filename}")
+                print(f"INFO: export to file {filename}:{status}")
             stls.append((j,filename))
         return stls
 
@@ -302,7 +344,7 @@ class Assembly:
             #construct stl_file names from stp_files
             self.stl_files=[str(pl.Path(f).with_suffix('.stl')) for f in self.stp_files]
 
-        # exports the reactor solid as a separate stl files
+        # exports the assembly solid as a separate stl files
         if len(filename) != len(self.entities):
             msg = (
                 f"The Reactor contains {len(self.shapes_and_components)} "
@@ -341,20 +383,36 @@ class Assembly:
             stl_list=self.gmsh_export_stls()
             if(heal):
                 stl_list=self.heal_stls(stl_list)
+
+            # now we have the gmsh imported geometry. This order may be different from the one
+            # that is held by the cq.entities. We should rerun the similarity filter.
+            brep_volume_dimtags=gmsh.model.occ.getEntities(3)
+            tag_idxs=[]
+            for dimtag in brep_volume_dimtags()
+              cms=gmsh.model.get_center_of_mass(dimtag[0],dimtag[1])
+              bb=gmsh.model.get_bounding_box(dimtag[0],dimtag[1])
+              vol=gmsh.model.get_mass(dimtag[0],dimtag[1])
+              j=idx_similar(self.entities,cms,bb,vol)
+              tag_idx.append(j)
+
             #add the material tags to the stl_list
             stl_tagged=[]
-            for (stl,e) in zip(stl_list,self.entities):
+            for (stl,e) in zip(stl_list,self.entities[i for i in tag_idx if !=-1] ):
                 stl_tagged.append((stl[0],stl[1],e.tag))
             self.stl2h5m(stl_tagged,h5m_file=h5m_filename)
         elif(backend=="stl"):
             if (self.verbose):
-                print(f'INFO: Using backend {stl} with parameters tolerance={tol} and angular tolerance={ang_tol}')
+                print(f'INFO: Using backend {backend} with parameters tolerance={stl_tol} and angular tolerance={stl_ang_tol}')
             stl_list=self.cq_export_stls(tolerance=stl_tol,angular_tolerance=stl_ang_tol)
             if(heal):
                 stl_list=self.heal_stls(stl_list)
             stl_tagged=[]
             for (stl,e) in zip(stl_list,self.entities):
-                stl_tagged.append((stl[0],stl[1],e.tag))
+                try:
+                    stl_tagged.append((stl[0],stl[1],e.tag))
+                except:
+                    print("WARNING: list of material tags is exhausted. Tagging volume {stl[0]},{stl[1]} with \'vacuum\'")
+                    stl_tagged.append(st[0],stl[1],'vacuum')
             self.stl2h5m(stl_tagged,h5m_file=h5m_filename)
         else:
             print(f'ERROR: Unknown backend: {backend}')
@@ -552,6 +610,24 @@ class Assembly:
             #the merge surface returns a cq-compound object.
             self.merged = self.merge_surfaces()
             rval=self.merged.exportBrep(str(path_filename))
+            #the merging process may result in extra volumes.
+            #We need to make sure these are at the end of the list
+            #If not this results in a loss ofvolumes in the end.
+            print("INFO: reordering volumes")
+            print(self.merged)
+            idxs=[]
+            for solid in self.merged.Solids():
+              center=solid.Center()
+              bb=solid.BoundingBox()
+              vol=solid.Volume()
+              idx=idx_similar(self.entities,center,bb,vol)
+              idxs.append(idx)
+            #reorder
+
+            print(idxs)
+            ents=[self.entities[i] for i in idxs if i!=-1]
+            self.entities=ents
+
         return rval
 
     def merge_surfaces(self):
@@ -572,7 +648,7 @@ class Assembly:
                   bldr.AddArgument(shape.solid.val().wrapped)
               except:
                   bldr.AddArgument(shape.solid.wrapped)
-
+        bldr.SetParallelMode_s(True)
         bldr.SetNonDestructive(True)
 
         if(self.verbose>1):
@@ -589,6 +665,12 @@ class Assembly:
 
         return self.merged
 
+    def merge_two(self,solid1,solid2):
+        """ Checks two surfaces if their BB overlap. If so merge the two - and return a list of the results
+        """
+
+
+
     def gmsh_init(self,brep_fn="geometry.brep",default=False,samples=20, min_mesh_size=0.1, max_mesh_size=10, mesh_algorithm=1, threads=None):
         gmsh.initialize()
         if (self.verbose>1):
@@ -602,8 +684,6 @@ class Assembly:
           #do this by means of properties instead
           if(threads is not None):
             gmsh.option.setNumber("General.NumThreads",threads)
-          self.volumes = gmsh.model.occ.importShapes(brep_fn)
-          gmsh.model.occ.synchronize()
 
           gmsh.option.setNumber("Mesh.Algorithm", mesh_algorithm)
           gmsh.option.setNumber("Mesh.MeshSizeMin", min_mesh_size)
@@ -613,6 +693,8 @@ class Assembly:
           gmsh.option.setNumber("Mesh.MeshSizeFromPoints",0)
           gmsh.option.setNumber("Mesh.MeshSizeExtendFromBoundary", 0)
           gmsh.option.setNumber("Mesh.MeshSizeFromCurvature", samples)
+        self.volumes = gmsh.model.occ.importShapes(brep_fn)
+        gmsh.model.occ.synchronize()
 
     def gmsh_deinit(self):
         gmsh.finalize()
@@ -663,8 +745,11 @@ class Assembly:
            pg = gmsh.model.addPhysicalGroup(2,ents[1])
            ps = gmsh.model.setPhysicalName(2,pg,f'surfaces_on_volume_{vid}')
            filename=f'volume_{vid}.stl'
-           gmsh.write(filename)
-           stls.append((vid,filename))
+           try:
+              gmsh.write(filename)
+              stls.append((vid,filename))
+           except:
+              print(f'WARNING: Could not write volume {vid}. Skipping')
            gmsh.model.removePhysicalGroups([]) # remove group again
         return stls
 
